import os
import timeit
import unittest

import numpy as np
from scipy.sparse import csr_matrix

from sparselsh import LSH
from tests.base_test import LSHTestBase


class BenchmarkingTestCase(unittest.TestCase, LSHTestBase):
    """
    Do some simple benchmarking

    NOTE: This test only runs if the SPARSELSH_BENCH environment
    variable is set. You can run it like so:

        SPARSELSH_BENCH=1 pytest tests/test_benchmark.py
    """
    def rand_sparse(self, m, n, sparsity=0.98):
        """
        Build a deterministc random sparse matrix. It's important
        that this is deterministic for reproducibility.
        """
        print("Building sparse random matrix...")
        _s = timeit.default_timer()
        np.random.seed(7)
        X_rnd = np.random.rand(m, n)
        np.putmask(X_rnd, X_rnd < sparsity, 0)
        X_sparse = csr_matrix(X_rnd)
        print(f"Done in {timeit.default_timer()-_s}s")
        return X_sparse

    @unittest.skipIf(not os.environ.get("SPARSELSH_BENCH"),
                     "Skipping benchmark")
    def test_query_performance(self):
        # small hashsize to create large buckets
        hashsize = 2
        # rows
        m = hashsize * 100_000
        # features
        n = 10
        # random sparse matrix
        X = self.rand_sparse(m, n)
        print("X:", X.__repr__())
        # make one huge sparse query matrix with lots of rows
        X_query = self.rand_sparse(10, n)
        print("X_query:", X_query.__repr__())

        num_hashtables = 1
        lsh_args = (hashsize, X.shape[1])
        lsh_kwargs = dict(
            num_hashtables=num_hashtables,
            storage_config={"dict": None}
        )
        lsh = LSH(
            *lsh_args, **lsh_kwargs
        )

        print("Starting index benchmark...")
        timing = timeit.timeit(lambda: lsh.index(X), number=1)
        print("Index time:", timing)

        print("Starting query benchmark...")
        timings = []
        for i in range(5):
            t = timeit.timeit(lambda: lsh.query(X_query), number=1)
            print(f"{i}: Query timing: {t}")
            timings.append(t)
        print("Query avg time:", np.mean(timings))


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
from __future__ import print_function

import argparse
import math
import operator
import re

import numpy as np
from scipy.sparse import csr_matrix

try:
    # Python 2
    import cPickle as pickle
except ImportError:
    # Python 3
    import pickle
try:
    # Python <= 2.7
    reduce
except NameError:
    # Python 3
    from functools import reduce

from sparselsh.lsh import LSH


def parse_args():
    desc = "An example tool that will build clusters using a LSH index from an input file containing one record per line. The output will be either a list of cluster groups or the within set sum of squared errors for the given hash size (enabling search for optimal hash-size when ran multiple times with different --hashsize params)."
    parser = argparse.ArgumentParser(description=desc)
    parser.add_argument(
        "corpus", type=str,
        help="Path to file, one record per line."
    )
    parser.add_argument(
        "--save-model", dest="save", action="store_true",
        help=(
            "Save model to specified file. By default, index is not "
            "saved."
        )
    )
    parser.add_argument(
        "--hashsize", type=int, default=256,
        help=(
            "Size of hash. Smaller sizes create fewer hash buckets and "
            "more dissimilar things will end up in the same bucket."
        )
    )
    parser.add_argument(
        "--num-tables", dest="tables", type=int, default=1,
        help=(
            "Number of hyperplanes to use. This doubles the potential "
            "memory usage but increases accuracy."
        )
    )
    parser.add_argument(
        "--output", type=str, choices=[
            "clusters", "wssse"
        ], default="clusters", help=(
            "What to output. `cluster`, default, prints out the cluster "
            "names and their items. `wssse` outputs variance inside "
            "clusters and their items. Useful for finding optimal "
            "hashsize."
        )
    )
    args = parser.parse_args()
    return args


def run(corpus_path, save_index=False, hashsize=128, output="clusters",
        num_tables=1):
    rawcorpus = None
    with open(corpus_path, "r") as f:
        rawcorpus = f.readlines()

    corpus = [re.sub("[^0-9a-z]", "", rc.lower()) for rc in rawcorpus]
    maxlen = max([len(x) for x in corpus])
    size = len(corpus)

    # break up text lines into fixed-length int matrix
    X = np.zeros((size, maxlen))
    for i in range(size):
        c = corpus[i]
        for j in range(len(c)):
            X[i][j] = ord(c[j])

    Xs = csr_matrix(X)
    lsh = LSH(
        hashsize, Xs.shape[1], num_hashtables=num_tables
    )
    lsh.index(Xs, extra_data=rawcorpus)

    if save_index:
        with open(save_index, "wb") as f:
            pickle.dump(lsh, f)

    if output == "clusters":
        t = lsh.hash_tables[0]
        print("Group #,Data")
        for k_ix, k in enumerate(t.keys()):
            vals = t.get_val(k)
            print()
            for val in vals:
                print(str(k_ix) + "," + re.sub("\r|\n", "", val[1]).strip())

    elif output == "wssse":
        cluster_keys = lsh.hash_tables[0].keys()

        def mean(cluster):
            cluster_matrices = map(lambda c: c[0], cluster)
            n = len(cluster_matrices)
            if n == 1:
                return cluster_matrices[0]
            return reduce(operator.add, cluster_matrices) / n

        def error(point, cluster_mean):
            return math.sqrt(
                sum([x**2 for x in (point - cluster_mean).toarray()[0]]))

        wssse = 0
        for key in cluster_keys:
            cluster = lsh.hash_tables[0].get_val(key)
            cluster_mean = mean(cluster)
            for point in cluster:
                # matrix is 0th item w/ extra_data 1st
                e = error(point[0], cluster_mean)
                wssse += e

        print(hashsize, wssse)


def cli():
    args = parse_args()
    save = args.save if args.save else False
    hashsize = args.hashsize
    num_tables = args.tables if args.tables else 1
    run(args.corpus, save_index=save, hashsize=hashsize, output=args.output,
        num_tables=num_tables)


if __name__ == "__main__":
    cli()



# thanks to S. Bhargav for submitting this test code
import unittest

from scipy.sparse import csr_matrix

from sparselsh import LSH


class IndexThenQueryTestCase(unittest.TestCase):
    """ Holding place for general regression tests for thing we
    find when indexing and then querying items.
    """
    def test_can_query_using_existing_row(self):
        X = csr_matrix([
            [3, 0, 0, 0, 0, 0, -1],
            [0, 1, 0, 0, 0, 0,  1],
            [1, 1, 1, 1, 1, 1,  1]
        ])

        # One class number for each input point
        y = ["0", "blah", "kill"]
        # I've changed the last 0 to a 1
        X_sim = csr_matrix([
            [1, 1, 1, 1, 1, 1, 1]
        ])

        lsh = LSH(4,
                  X.shape[1],
                  num_hashtables=1,
                  storage_config={"dict":None})

        for ix in range(X.shape[0]):
            x = X.getrow(ix)
            c = y[ix]
            lsh.index(x, extra_data=c)

        # find the point in X nearest to X_sim
        distance_functions = (
            None,
            "hamming", "euclidean", "true_euclidean",
            "cosine", "l1norm"
        )
        for dist in distance_functions:
            points = lsh.query(X_sim, distance_func=dist, num_results=1)

        self.assertEqual(type(points[0][0][0]), csr_matrix)
        truth = points[0][0][0].todense() == X_sim.todense()
        self.assertTrue(truth.all())


if __name__ == '__main__':
    unittest.main()



import unittest

from scipy.sparse import csr_matrix

from sparselsh import LSH
from tests.base_test import LSHTestBase


class MatrixIndexTestCase(unittest.TestCase, LSHTestBase):
    """
    Ensure we can pass rows and matrix to the index function
    and have it work the same. This tests both row format
    and matrix format, ensures the projected hash tables are
    the same and also the query results are the same.
    """

    def test_can_index_matrix_input(self):
        X = csr_matrix([
            [3, 0, 0, 0, 0, 0, -1],
            [0, 1, 0, 0, 0, 0,  1],
            [1, 1, 1, 1, 1, 1,  1]
        ])

        # One class number for each input point
        y = ["one", "two", "three"]

        # two comparisons
        X_sim = csr_matrix([
            [1, 1, 1, 1, 1, 1, 0],
            [1, 1, 1, 1, 1, 1, 1]
        ])

        # use same LSH args for both
        num_hashtables = 2
        lsh_args = (4, X.shape[1])
        lsh_kwargs = dict(
            num_hashtables=num_hashtables,
            storage_config={"dict": None}
        )

        # create LSH instances
        lsh_row = LSH(
            *lsh_args, **lsh_kwargs
        )
        lsh_matrix = LSH(
            *lsh_args, **lsh_kwargs
        )
        # sync uniform planes so hashes are the same
        uniform_planes = lsh_row._generate_uniform_planes()
        lsh_row.uniform_planes = uniform_planes
        lsh_matrix.uniform_planes = uniform_planes

        # add each item row-wise (old version)
        for ix in range(X.shape[0]):
            x = X.getrow(ix)
            c = y[ix]
            lsh_row.index(x, extra_data=c)

        # then index again as a matrix
        lsh_matrix.index(X, extra_data=y)

        # make sure the hash tables are the same
        for table_index, hash_table1 in enumerate(lsh_row.hash_tables):
            hash_table2 = lsh_matrix.hash_tables[table_index]
            keys1 = hash_table1.keys()
            keys2 = hash_table2.keys()
            self.assertEqual(keys1, keys2)
            for key in keys1:
                values1 = hash_table1.get_val(key)
                values2 = hash_table2.get_val(key)
                self.assertEqual(
                    len(values1), len(values2),
                    f"Table keys '{key}' have different numbers of values"
                )
                for result_ix, (val1, extra1) in enumerate(values1):
                    val2, extra2 = values2[result_ix]
                    eq = val1.todense() == val2.todense()
                    self.assertTrue(eq.all, "Resulting points mismatch!")
                    self.assertEqual(extra1, extra2, "Extra data mismatch!")

        # returns list of results per input point
        query_results = lsh_row.query(
            X_sim, num_results=2
        )
        print("* query_results", query_results)
        self.assertLessEqual(
            X_sim.shape[0], len(query_results),
            "Query length should match input points"
        )
        first_query_result = query_results[0]
        print("first_query_result", first_query_result)
        self.assertEqual(
            len(first_query_result), 2,
            "Results data should have a max of two results"
        )
        # first_query_result is tuple of N results, N extra data and, if
        # there are multiple results, a lists of distances for each.
        print("first_query_result[0]", first_query_result[0])
        (point1, extra1), dist1 = first_query_result[0]

        query_results = lsh_matrix.query(
            X_sim, num_results=2
        )
        first_query_result = query_results[0]
        (point2, extra2), dist2 = first_query_result[0]
        print("point2", point2, "extra2", extra2)

        print("points1 & 2", point1, point2)
        self.assertSparseEqual(point1, point2)
        self.assertEqual(extra1, extra2)
        self.assertEqual(dist1, dist2)

    def test_with_inp_matrix_with_dist_threshold(self):
        X = csr_matrix([
            [3, 0, 0, 0, 0, 0, -1],
            [0, 1, 0, 0, 0, 0,  1],
            [1, 1, 1, 1, 1, 1,  1]
        ])

        # One class number for each input point
        y = ["one", "two", "three"]

        # I've changed the last 1 to a 0
        X_sim = csr_matrix([
            [1, 1, 1, 1, 1, 1, 0],
            [1, 1, 1, 1, 1, 0, 0]
        ])

        lsh_args = (4, X.shape[1])
        lsh_kwargs = dict(
            num_hashtables=2,
            storage_config={"dict": None}
        )
        lsh = LSH(
            *lsh_args, **lsh_kwargs
        )
        self.load_planes(lsh)
        lsh.index(X, extra_data=y)

        results_all = lsh.query(X_sim, distance_func="cosine",
                                dist_threshold=1.0)
        print("results_all", results_all)
        self.assertEqual(
            len(results_all[0]), X.shape[0],
            "Incorrect number of results for dist_threshold < 1.0"
        )

        results_some = lsh.query(X_sim, distance_func="cosine",
                                 dist_threshold=0.5)
        print("results_some", results_some)
        self.assertEqual(
            len(results_some[0]), 2,
            "Incorrect number of results for dist_threshold < 0.5"
        )

        results_none = lsh.query(X_sim, distance_func="cosine",
                                 dist_threshold=0.00001)
        print("results_none", results_none)
        self.assertEqual(
            len(results_none[0]), 0,
            "Incorrect number of results for dist_threshold near zero"
        )


if __name__ == '__main__':
    unittest.main()



""" This file implements are very crude, brute-force test function
that makes sure that all storage backends work (don't crash) with
all distance functions. This will write to disk, etc. Run from the
SparseLSH root directory.
"""
import sys
import unittest

from scipy.sparse import csr_matrix

from sparselsh import LSH


try:
    import redis
except ImportError:
    redis = None

try:
    import bsddb
except ImportError:
    bsddb = None

try:
    import leveldb
except ImportError:
    leveldb = None


X = csr_matrix([
    [0, 0, 0, 0, 0, 0,  1],
    [3, 0, 0, 0, 0, 0, -1],
    [0, 1, 0, 0, 0, 0,  1],
    [0, 0, 0, 9, 0, 0,  1],
    [0, 0, 0, 0, 0, 0,  0],
    [4, 0, 0, 0, 0, 0,  0],
    [0, 0, 0, 0, 0, 0,  1],
    [0, 0, 2, 0, 0, 0,  1],
    [0, 0, 0, 0, 0, 0,  0],
    [1, 1, 1, 1, 1, 1,  1]
])

X_sim = csr_matrix([
    [1, 1, 1, 1, 1, 1, 0]
])
Y_sim = [
    10
]

# Simulate extra-data / classes
y = [
    'one',
    'two',
    'three',
    'four',
    'zero',
    'zero',
    'seven',
    'eight',
    'nine',
    'last',
]


def lsh_instance(bits, dimensions, hashtables, storage_type):
    """ Build an lsh instance using specified settings & storage method.

    Params:
        bits: n bitlength of hash key
        dimentions: n features in each input point
        hashtables: number of parallel hashtables to construct
        storage_type: type of storage backend. one of:
            'berkeleydb', 'leveldb', 'redis', 'dict'
    """
    if storage_type == 'berkeleydb':
        if sys.version_info[0] < 3:
            conf = {'berkeleydb':{'filename': './db'}}
        else:
            raise NotImplementedError('BerkeleyDB not supported in Python3')
    elif storage_type == 'leveldb':
        conf = {'leveldb':{'db': 'ldb'}}
    elif storage_type == 'redis':
        conf = {'redis': {'host': '127.0.0.1', 'port': 6379}}
    else:
        conf = {'dict': None}

    lsh = LSH(bits,
              dimensions,
              num_hashtables=hashtables,
              storage_config=conf)
    return lsh


def build_index(lsh, X, y):
    for ix in range(X.shape[0]):
        x = X.getrow(ix)
        c = None
        if y:
            c = y[ix]
        lsh.index(x, extra_data=c)


def query_index(lsh, x):
    similar = lsh.query(x, num_results=1)
    #print "SIMILAR", similar
    return similar


def run_lsh(bits, dimensions, hashtables, storage_type, distance_fn):
    lsh = lsh_instance(bits, dimensions, hashtables, storage_type)
    build_index(lsh, X, y)
    # [(<1x7 sparse matrix of type '<class 'numpy.int64'>'>, 'last', array([1.]))]
    similar_points = query_index(lsh, X_sim.getrow(0))
    print("similar_points", similar_points)
    # for s_point in similar_points:
    #     if not s_point:
    #         continue

    #     # first result from first plane
    #     print("s_point", s_point)
    #     point = s_point[0]
    #     print("point", point)
    #     extra_data = s_point[1]
    #     similarity = None
    #     if len(s_point) == 3:
    #         similarity = s_point[2]
    #     print("extra_data", extra_data)
    #     print("similarity", similarity)

    #     # print "Found similar point", point.todense()
    #     # if extra_data:
    #     #     print "with data", extra_data
    #     # print "similarity", similarity
    #     errmsg = "Similar point returned is invalid (not a sparse csr mx)"
    #     assert(type(point) == csr_matrix), errmsg

    for result_ix, s_point in enumerate(similar_points):
        print("result_ix:", result_ix, "s_point:", s_point)
        point = s_point[0][0]
        print("point:", point.todense())
        extra_data = s_point[0][1] if len(s_point[0]) == 2 else None
        print("extra_data:", extra_data)
        similarity = s_point[1]
        # print "Found similar point", point.todense()
        # if extra_data:
        #     print "with data", extra_data
        # print "similarity", similarity
        print("similarity:", similarity)
        errmsg = "Similar point returned is invalid (not a sparse csr mx)"
        assert(type(point) == csr_matrix), errmsg




class TestCase(unittest.TestCase):
    def test_e2e_run_lsh(self):
        # Run this whole thing this many times to tease our
        # any errors caused by plane randomization
        n_runs = 5
        # Settings
        bits = 4
        dimensions = X.shape[1]
        hashtables = [1, 2, 3]
        storage_types = ["dict"]
        # if redis is not None:
        #     storage_types.append("redis")
        if bsddb is not None:
            storage_types.append("berkeleydb")
        if leveldb is not None:
            storage_types.append("leveldb")

        if sys.version_info[0] == 2:
            storage_types.append("berkeleydb")
        distance_functions = (
            "hamming", "euclidean", "true_euclidean",
            "centred_euclidean", "cosine", "l1norm"
        )
        for _ in range(n_runs):
            for h in hashtables:
                for s in storage_types:
                    print("Testing with backend", s)
                    for d in distance_functions:
                        print("Testing distance function", d)
                        run_lsh(bits, dimensions, h, s, d)


if __name__ == "__main__":
    unittest.main()



import unittest

from scipy.sparse import csr_matrix

from sparselsh import LSH
from tests.base_test import LSHTestBase


class NoExtraDataTestCase(unittest.TestCase, LSHTestBase):
    """
    Some basic tests without using extra data
    """

    def test_can_index_matrix_input(self):
        X = csr_matrix([
            [3, 0, 0, 0, 0, 0, -1],
            [0, 1, 0, 0, 0, 0,  1],
            [1, 1, 1, 1, 1, 1,  1]
        ])

        # I've changed the last 1 to a 0
        X_sim = csr_matrix([
            [1, 1, 1, 1, 1, 1, 0],
            [1, 1, 1, 1, 1, 1, 0]
        ])

        num_hashtables = 2
        lsh_args = (4, X.shape[1])
        lsh_kwargs = dict(
            num_hashtables=num_hashtables,
            storage_config={"dict": None}
        )
        lsh = LSH(
            *lsh_args, **lsh_kwargs
        )
        self.load_planes(lsh)
        lsh.index(X)

        # returns list of results per input point
        num_results = 2
        query_results = lsh.query(
            X_sim, num_results=num_results
        )
        print("query_results", query_results)
        self.assertEqual(
            len(query_results), X_sim.shape[0],
            "Query results len should equal number of query points"
        )
        first_point_result = query_results[0]
        print("first_point_result", len(first_point_result),
              first_point_result)

        self.assertTrue(
            len(first_point_result) <= num_results,
            "Query results should have correct number of results"
        )

        print("first_point_result[0]", first_point_result[0])
        self.assertEqual(
            len(first_point_result[0]), 2,
            "Result should have two items: result data, dist"
        )

        s_point, similarity = first_point_result[0]
        self.assertEqual(
            len(s_point), 1,
            "Result tuple length should only have one item"
        )
        point = s_point[0]

        self.assertSparseEqual(point, csr_matrix(X.getrow(2)))
        self.assertTrue(similarity is not None)

    def test_with_distance_threshold(self):
        X = csr_matrix([
            [3, 0, 0, 0, 0, 0, -1],
            [0, 1, 0, 0, 0, 0,  1],
            [1, 1, 1, 1, 1, 1,  1]
        ])

        # I've changed the last 1 to a 0
        X_sim = csr_matrix([
            [1, 1, 1, 1, 1, 1, 0],
            [1, 1, 1, 1, 1, 0, 0]]
        )

        lsh_args = (4, X.shape[1])
        lsh_kwargs = dict(
            num_hashtables=2,
            storage_config={"dict": None}
        )
        lsh = LSH(
            *lsh_args, **lsh_kwargs
        )
        self.load_planes(lsh)
        lsh.index(X)

        results_all = lsh.query(X_sim, distance_func="cosine",
                                dist_threshold=1.0)
        print("results_all", results_all)
        self.assertEqual(
            len(results_all[0]), X.shape[0],
            "Incorrect number of results for dist_threshold < 1.0"
        )

        results_some = lsh.query(X_sim, distance_func="cosine",
                                dist_threshold=0.5)
        print("results_some", results_some)
        self.assertEqual(
            len(results_some[0]), 2,
            "Incorrect number of results for dist_threshold < 0.5"
        )

        results_none = lsh.query(X_sim, distance_func="cosine",
                                dist_threshold=0.00001)
        print("results_none", results_none)
        self.assertEqual(
            len(results_none[0]), 0,
            "Incorrect number of results for dist_threshold near zero"
        )


if __name__ == '__main__':
    unittest.main()



from __future__ import print_function

import pickle
import sys

try:
    import redis
except ImportError:
    redis = None

try:
    import bsddb
except ImportError:
    bsddb = None

try:
    import leveldb
except ImportError:
    leveldb = None

try:
    import cPickle as pickle
except ImportError:
    import pickle


def serialize(data):
    return pickle.dumps(data, protocol=2)
def deserialize(data):
    return pickle.loads(data)


__all__ = ['storage']


def storage(storage_config, index):
    """ Given the configuration for storage and the index, return the
    configured storage instance.
    """
    if 'dict' in storage_config:
        return InMemoryStorage(storage_config['dict'])
    elif 'redis' in storage_config:
        storage_config['redis']['db'] = index
        return RedisStorage(storage_config['redis'])
    elif 'berkeleydb' in storage_config:
        return BerkeleyDBStorage(storage_config['berkeleydb'])
    elif 'leveldb' in storage_config:
        return LevelDBStorage(storage_config['leveldb'])
    else:
        raise ValueError("Only in-memory dictionary, berkeleydb, leveldb, and redis are supported.")


class BaseStorage(object):
    def __init__(self, config):
        """ An abstract class used as an adapter for storages. """
        raise NotImplementedError

    def serialize(self, data):
        return serialize(data)

    def deserialize(self, data):
        return deserialize(data)

    def keys(self):
        """ Returns a list of binary hashes that are used as dict keys. """
        raise NotImplementedError

    def set_val(self, key, val):
        """ Set `val` at `key`, note that the `val` must be a string. """
        raise NotImplementedError

    def get_val(self, key):
        """ Return `val` at `key`, note that the `val` must be a string. """
        raise NotImplementedError

    def append_val(self, key, val):
        """ Append `val` to the list stored at `key`.

        If the key is not yet present in storage, create a list with `val` at
        `key`.
        """
        raise NotImplementedError

    def get_list(self, key):
        """ Returns a list stored in storage at `key`.

        This method should return a list of values stored at `key`. `[]` should
        be returned if the list is empty or if `key` is not present in storage.
        """
        raise NotImplementedError


class InMemoryStorage(BaseStorage):
    def __init__(self, config):
        self.name = 'dict'
        self.storage = dict()

    def keys(self):
        return list(self.storage.keys())

    def set_val(self, key, val):
        self.storage[key] = val

    def get_val(self, key):
        return self.storage[key]

    def append_val(self, key, val):
        self.storage.setdefault(key, []).append(val)

    def get_list(self, key):
        return self.storage.get(key, [])


class RedisStorage(BaseStorage):
    def __init__(self, config):
        if not redis:
            raise ImportError("redis-py is required to use Redis as storage.")
        self.name = 'redis'
        self.storage = redis.StrictRedis(**config)

    def keys(self, pattern="*"):
        return self.storage.keys(pattern)

    def set_val(self, key, val):
        self.storage.set(key, val)

    def get_val(self, key):
        return self.storage.get(key)

    def append_val(self, key, val):
        self.storage.rpush(key, serialize(val))

    def get_list(self, key):
        # TODO: find a better way to do this
        values = self.storage.lrange(key, 0, -1)
        return [deserialize(v)for v in values]


class BerkeleyDBStorage(BaseStorage):
    def __init__(self, config):
        if 'filename' not in config:
            raise ValueError("You must supply a 'filename' in your config")
        self.storage = bsddb.hashopen(config['filename'])

    def __exit__(self, type, value, traceback):
        self.storage.sync()

    def keys(self):
        return list(self.storage.keys())

    def set_val(self, key, val):
        self.storage[key] = self.serialize(val)

    def get_val(self, key):
        return self.storage[key]

    def append_val(self, key, val):
        try:
            current = self.deserialize(self.storage[key])
        except KeyError:
            current = []

        # update new list
        current.append(val)
        self.storage[key] = self.serialize(current)

    def get_list(self, key):
        try:
            return self.deserialize(self.storage[key])
        except KeyError:
            return []


class LevelDBStorage(BaseStorage):
    def __init__(self, config):
        if not leveldb:
            raise ImportError("leveldb is required to use LevelDB as storage.")
        if 'db' not in config:
            raise ValueError("You must specify LevelDB filename as 'db' in your config")
        self.storage = leveldb.LevelDB(config['db'])

    def keys(self):
        return self.storage.RangeIter(include_value=False)

    def set_val(self, key, val):
        self.storage.Put(key, self.serialize(val))

    def get_val(self, key):
        return self.serialize(self.storage.Get(key))

    def append_val(self, key, val):
        # If a key doesn't exist, leveldb will throw KeyError
        current = []
        try:
            item = self.storage.Get(key)
            current = self.deserialize()
        except KeyError:
            pass
        except TypeError:
            # here, we have python3, which has a different manner for
            # interacting with leveldb
            if type(key) == str and sys.version_info[0] == 3:
                return self.append_val(bytes(key, 'UTF-8'), val)

        # update new list
        current.append(val)
        self.storage.Put(key, self.serialize(current))

    def get_list(self, key):
        try:
            k = self.storage.Get(key)
            return self.deserialize(k)
        except KeyError:
            return []



# This module is part of lshash and is released under
# the MIT License: http://www.opensource.org/licenses/mit-license.php

__title__ = 'sparselsh'
__author__ = 'Brandon Roberts (brandon@bxroberts.org)'
__license__ = 'MIT'
__version__ = '2.1.1'

from sparselsh.lsh import LSH



import setuptools


with open("README.md", "r") as fh:
    long_description = fh.read()


setuptools.setup(
    name='sparselsh',
    version='2.1.1',
    author='Brandon Roberts',
    author_email='brandon@bxroberts.org',
    description='A locality sensitive hashing library with an emphasis on large, sparse datasets.',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='https://github.com/brandonrobertz/sparselsh',
    download_url='https://github.com/brandonrobertz/SparseLSH/releases/tag/v2.1.1',
    keywords=['clustering', 'sparse', 'lsh'],
    packages=setuptools.find_packages(),
    install_requires=[
        'numpy>=1.18.6,<2.0',
        'scipy>=1.0.0,<2.0',
        'scikit-learn>=0.24.0,<2.0'
    ],
    extras_require={
        "test": ["pytest"],
        "redis": ["redis>=2.10.1,<3.0"]
    },
    entry_points="""
        [console_scripts]
        sparselsh=sparselsh.cli:cli
    """,
    classifiers=[
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Topic :: Software Development :: Libraries',
        ],
)



#!/usr/bin/env python
from __future__ import print_function

import os
from operator import itemgetter

import numpy as np
from scipy.sparse import csr_matrix, issparse, vstack
from scipy.spatial.distance import hamming
from sklearn.metrics.pairwise import cosine_distances

from .storage import storage, serialize, deserialize


class LSH(object):
    """ LSH implments locality sensitive hashing using random projection for
    input vectors of dimension `input_dim`.

    Attributes:

    :param hash_size:
        The length of the resulting binary hash in integer. E.g., 32 means the
        resulting binary hash will be 32-bit long.
    :param input_dim:
        The dimension of the input vector. This can be found in your sparse
        matrix by checking the .shape attribute of your matrix. I.E.,
            `csr_dataset.shape[1]`
    :param num_hashtables:
        (optional) The number of hash tables used for multiple look-ups.
        Increasing the number of hashtables increases the probability of
        a hash collision of similar documents, but it also increases the
        amount of work needed to add points.
    :param storage_config:
        (optional) A dictionary of the form `{backend_name: config}` where
        `backend_name` is the either `dict`, `berkeleydb`, `leveldb` or
        `redis`. `config` is the configuration used by the backend.
        Example configs for each type are as follows:
        `In-Memory Python Dictionary`:
            {"dict": None} # Takes no options
        `Redis`:
            `{"redis": {"host": hostname, "port": port_num}}`
            Where `hostname` is normally `localhost` and `port` is normally 6379.
        `LevelDB`:
            {'leveldb':{'db': 'ldb'}}
            Where 'db' specifies the directory to store the LevelDB database.
        `Berkeley DB`:
            {'berkeleydb':{'filename': './db'}}
            Where 'filename' is the location of the database file.
        NOTE: Both Redis and Dict are in-memory. Keep this in mind when
        selecting a storage backend.
    :param matrices_filename:
        (optional) Specify the path to the compressed numpy file ending with
        extension `.npz`, where the uniform random planes are stored, or to be
        stored if the file does not exist yet.
    :param overwrite:
        (optional) Whether to overwrite the matrices file if it already exist.
        This needs to be True if the input dimensions or number of hashtables
        change.
    """

    def __init__(self, hash_size, input_dim, num_hashtables=1,
                 storage_config=None, matrices_filename=None, overwrite=False):

        assert hash_size >= 1, "hash_size must be a positive integer"
        assert input_dim >= 1, "input_dim must be a positive integer"
        assert num_hashtables >= 1, "num_hashtables must be a positive integer"

        self.hash_size = hash_size
        self.input_dim = input_dim
        self.num_hashtables = num_hashtables

        if storage_config is None:
            storage_config = {'dict': None}
        self.storage_config = storage_config

        if matrices_filename and not matrices_filename.endswith('.npz'):
            raise ValueError("The specified file name must end with .npz")
        self.matrices_filename = matrices_filename
        self.overwrite = overwrite

        self._init_uniform_planes()
        self._init_hashtables()

    def _init_uniform_planes(self):
        """ Initialize uniform planes used to calculate the hashes

        if file `self.matrices_filename` exist and `self.overwrite` is
        selected, save the uniform planes to the specified file.

        if file `self.matrices_filename` exist and `self.overwrite` is not
        selected, load the matrix with `np.load`.

        if file `self.matrices_filename` does not exist and regardless of
        `self.overwrite`, only set `self.uniform_planes`.
        """

        if "uniform_planes" in self.__dict__:
            return

        if self.matrices_filename:
            file_exist = os.path.isfile(self.matrices_filename)
            if file_exist and not self.overwrite:
                try:
                    # TODO: load sparse file
                    npzfiles = np.load(self.matrices_filename)
                except IOError:
                    print("Cannot load specified file as a numpy array")
                    raise
                else:
                    npzfiles = sorted(list(npzfiles.items()), key=lambda x: x[0])
                    # TODO: to sparse
                    self.uniform_planes = [t[1] for t in npzfiles]
            else:
                self.uniform_planes = [self._generate_uniform_planes()
                                       for _ in range(self.num_hashtables)]
                try:
                    np.savez_compressed(self.matrices_filename,
                                        *self.uniform_planes)
                except IOError:
                    print("IOError when saving matrices to specificed path")
                    raise
        else:
            self.uniform_planes = [self._generate_uniform_planes()
                                   for _ in range(self.num_hashtables)]

    def _init_hashtables(self):
        """ Initialize the hash tables such that each record will be in the
        form of "[storage1, storage2, ...]" """

        self.hash_tables = [storage(self.storage_config, i)
                            for i in range(self.num_hashtables)]

    def _generate_uniform_planes(self):
        """ Generate uniformly distributed hyperplanes and return it as a 2D
        numpy array.
        """
        dense_planes = np.random.randn(self.hash_size, self.input_dim)
        return csr_matrix(dense_planes)

    def _hash(self, planes, input_points):
        """ Generates the binary hashes for `input_points` and returns them.

        :param planes:
            The planes are random uniform planes with a dimension of
            `hash_size` x `input_dim`.
        :param input_points:
            A scipy sparse matrix that contains only numbers.
            The dimension needs to be N x `input_dim`, N>0.
        """
        try:
            planes = planes.transpose()
            projections = input_points.dot(planes)
            signs = (projections > 0)

        except TypeError as e:
            print("""The input point needs to be an array-like object with
                  numbers only elements""")
            raise
        except ValueError as e:
            print(("""The input point needs to be of the same dimension as
                  `input_dim` when initializing this LSH instance""", e))
            raise
        else:
            return np.packbits(signs.toarray(), axis=-1)

    def _as_np_array(self, serial_or_sparse):
        """ Takes either a serialized data structure, a sparse matrix, or tuple
        that has the original input points stored, and returns the original
        input point (a 1 x N sparse matrix).
        """
        # if we get a plain sparse matrix, return it (it's the point itself)
        if issparse(serial_or_sparse):
            return serial_or_sparse

        # here we have a serialized pickle object
        if isinstance(serial_or_sparse, str):
            try:
                deserial = deserialize(serial_or_sparse)
            except TypeError:
                print("The value stored is not deserializable")
                raise
        else:
            # If extra_data exists, `tuple` is the entire
            # (point:sparse, extra_daa). Otherwise (i.e., extra_data=None),
            # return the point stored as a tuple
            deserial = serial_or_sparse

        # if we deserialized it, we might have the sparse now
        if issparse(deserial):
            return deserial

        if isinstance(deserial[0], tuple):
            # extra data was supplied, return point
            return tuple[0]

        elif isinstance(deserial, (tuple, list)):
            try:
                return deserial[0]
            except ValueError as e:
                print(("The input needs to be an array-like object", e))
                raise
        else:
            raise TypeError("The input data is not supported")

    def _bytes_string_to_array(self, hash_key):
        """ Takes a hash key (bytes string) and turn it
        into a numpy matrix we can do calculations with.

        :param hash_key
        """
        return np.array(list(hash_key))

    def index(self, input_points, extra_data=None):
        """ Index input points by adding them to the selected storage.

        If `extra_data` is provided, it will become the value of the dictionary
        {input_point: extra_data}, which in turn will become the value of the
        hash table.

        :param input_points:
            A sparse CSR matrix. The dimension needs to be N x `input_dim`, N>0.
        :param extra_data:
            (optional) A list of values to associate with the points. Commonly
            this is a target/class-value of some type.
        """

        assert issparse(input_points), "input_points needs to be a sparse matrix"
        assert input_points.shape[1] == self.input_dim, "input_points wrong 2nd dimension"
        assert input_points.shape[0] == 1 or (input_points.shape[0] > 1 and \
               (extra_data is None or len(extra_data) == input_points.shape[0])), \
               "input_points dimension needs to match extra data size"

        for i, table in enumerate(self.hash_tables):
            keys = self._hash(self.uniform_planes[i], input_points)
            # NOTE: there was a bug with 0-equal extra_data
            # we need to allow blank extra_data if it's provided
            if extra_data is not None:
                # NOTE: needs to be tuple so it's set-hashable
                for j in range(keys.shape[0]):
                    # only split up extra_data if our input shape is greater
                    # than one. in that case we assume extra_data length is
                    # the same as length of the input points. not doing this
                    # causes the extra_data to be truncated to a single char
                    extra = extra_data[j] if input_points.shape[0] > 1 else extra_data
                    value = (input_points[j], extra)
                    table.append_val(keys[j].tobytes(), value)
            else:
                for j in range(keys.shape[0]):
                    value = (input_points[j],)
                    table.append_val(keys[j].tobytes(), value)

    def query(self, query_points, distance_func=None, dist_threshold=None,
              num_results=None):
        """ Takes `query_points` which is a sparse CSR matrix of N x `input_dim`,
        returns `num_results` of results as a list of tuples that are ranked
        based on the supplied metric function `distance_func`. The exact return
        value dimensions depends on if query_points is 1 or more dimensional.

        Example query return with 1-dimensional (single row) query_points:

            ```
            # Dimension 0: Results
            [
                # Dimension 1: Data & distance
                (
                    # Tuple: record, label
                    (<1x7 sparse matrix of type ...>, 'three'),
                    # Distance score
                    1.0
                ),
                (
                    (<1x7 sparse matrix of type ...>, 'three'),
                    1.0
                )
            ]
            ```

        Example query return with 2 query_points:

            ```
            # Dimension 0: Query points
            [
                # Dimension 1: Results
                [
                    # Dimension 2: Data & distance
                    (
                        # Tuple: record, label
                        (<1x7 sparse matrix of type ...>, 'three'),
                        # Distance score
                        1.0
                    ),
                    (
                        (<1x7 sparse matrix of type ...>, 'three'),
                        1.0
                    )
                ]
            ]
            ```

        :param query_points:
            A sparse CSR matrix. The dimension needs to be N x `input_dim`, N>0.
            Used by :meth:`._hash`.
        :param num_results:
            (optional) Integer, specifies the max amount of results to be
            returned. If not specified all candidates will be returned as a
            list in ranked order.
            NOTE: You do not save processing by limiting the results. Currently,
            a similarity ranking and sort is done on all items in the hashtable.
        :param distance_func:
            (optional) The distance function to be used. Currently it needs to
            be one of ("hamming", "euclidean", "true_euclidean",
            "cosine", "l1norm"). By default "euclidean"
            will used.
        :param dist_threshold:
            (optional) Its type and value depend on the chosen distance function.
            Specifies the distance threshold below which we accept a match. If not
            specified then any distance is accepted.
        """
        assert issparse(query_points), "query_points needs to be sparse"
        assert query_points.shape[1] == self.input_dim, "query_points wrong 2nd dimension"
        assert num_results is None or num_results >= 1, "num_results needs to be a positive integer"

        if distance_func is None or distance_func == "euclidean":
            d_func = LSH.euclidean_dist_square
        elif distance_func == "true_euclidean":
            d_func = LSH.euclidean_dist
        elif distance_func == "cosine":
            d_func = LSH.cosine_dist
        elif distance_func == "l1norm":
            d_func = LSH.l1norm_dist
        elif distance_func == "hamming":
            d_func = LSH.hamming_dist
        else:
            raise ValueError(
                "The distance function %s is invalid." % distance_func
            )
        if dist_threshold and (dist_threshold <= 0 or 
                               (distance_func == "cosine" and dist_threshold > 1.0)):
            raise ValueError(
                "The distance threshold %s is invalid." % dist_threshold
            )

        # Create a list of lists of candidate neighbors
        # NOTE: Currently this only does exact matching on hash key, the
        # previous version also got the 2 most simlilar hashes and included
        # the contents of those as well. Not sure if this should actually
        # be done since we can change hash size or add more tables to get
        # better accuracy
        candidates = []
        for i, table in enumerate(self.hash_tables):
            # get hashes of query points for the specific plane
            keys = self._hash(self.uniform_planes[i], query_points)
            for j in range(keys.shape[0]):
                # TODO: go through each hashkey in table and do the following
                # if the key is more similar to hashkey than hamming dist < 2

                # Create a sublist of candidate neighbors for each query point
                if len(candidates) <= j:
                    candidates.append([])
                new_candidates = table.get_list(keys[j].tobytes())
                if new_candidates is not None and len(new_candidates) > 0:
                    candidates[j].extend(new_candidates)

        # Create a ranked list of lists of candidate neighbors
        ranked_candidates = []

        # for each query point ...
        # Create a sublist of ranked candidate neighbors for each query point
        for j in range(query_points.shape[0]):
            point_results = []
            # hash candidates from above for jth query point
            row_candidates = candidates[j]

            # Remove extra info from candidates to convert them into a matrix
            cands = []
            extra_datas = []
            for row in row_candidates:
                cands.append(row[0])
                if len(row) == 2:
                    extra_datas.append(row[1])

            if not cands:
                ranked_candidates.append(point_results)
                continue

            cand_csr = vstack(cands)
            distances = d_func(query_points[j], cand_csr)
            if dist_threshold is not None:
                accepted = np.unique(np.where(distances < dist_threshold)[0])
                cand_csr = cand_csr[accepted, :]
                distances = distances[accepted]

            # Rank candidates by distance function, this has to
            # support having empty cand_csr and distances
            indices = np.argsort(np.array(distances))
            neighbors_sorted = cand_csr[indices]
            dists_sorted = distances[indices]

            # if we have extra data
            if extra_datas:
                # Sort extra_data by ranked distances
                try:
                    extra_data_sorted = itemgetter(*list(indices))(extra_datas)
                # we have no results, so no extra_datas
                except TypeError:
                    extra_data_sorted = []

                neigh_extra_tuples = zip(neighbors_sorted, extra_data_sorted)
                for ix, (neigh, ext) in enumerate(neigh_extra_tuples):
                    if num_results is not None and ix >= num_results:
                        break
                    dist = dists_sorted[ix]
                    point_results.append(((neigh, ext), dist))

            else:
                for ix, neigh in enumerate(neighbors_sorted):
                    if num_results is not None and ix >= num_results:
                        break
                    dist = dists_sorted[ix]
                    point_results.append(((neigh,), dist))

            ranked_candidates.append(point_results)

        if query_points.shape[0] == 1:
            # Backwards compat fix
            return ranked_candidates[0]

        return ranked_candidates

    ### distance functions
    @staticmethod
    def hamming_dist(x, Y):
        dists = np.zeros(Y.shape[0])
        for ix, y in enumerate(Y):
            dists[ix] = hamming(x, y)
        return dists

    @staticmethod
    def euclidean_dist(x, Y):
        # repeat x as many times as the number of rows in Y
        xx = csr_matrix(np.ones([Y.shape[0], 1]) * x)
        diff = Y - xx
        dists = np.sqrt(diff.dot(diff.T).diagonal()).reshape((1,-1))
        return dists[0]

    @staticmethod
    def euclidean_dist_square(x, Y):
        # repeat x as many times as the number of rows in Y
        xx = csr_matrix(np.ones([Y.shape[0], 1]) * x)
        diff = Y - xx
        if diff.nnz == 0:
            dists = np.zeros((1, Y.shape[0]))
        else:
            if Y.shape[0] > 1:
                dists = diff.dot(diff.T).diagonal().reshape((1,-1))
            else:
                dists = diff.dot(diff.T).toarray()
        return dists[0]

    @staticmethod
    def l1norm_dist(x, Y):
        # repeat x as many times as the number of rows in Y
        xx = csr_matrix(np.ones([Y.shape[0], 1]) * x)
        dists = np.asarray(abs(Y - xx).sum(axis=1).reshape((1,-1)))
        return dists[0]

    @staticmethod
    def cosine_dist(x, Y):
        dists = cosine_distances(x, Y)
        return dists[0]
